
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
from unittest import mock


import sushy
from sushy.resources.fabric import endpoint
from sushy.resources.fabric import fabric
from sushy.tests.unit import base


class FabricTestCase(base.TestCase):

    def setUp(self):
        super().setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/unit/json_samples/fabric.json') as f:
            self.json_doc = json.load(f)

        self.conn.get.return_value.json.return_value = self.json_doc

        self.fabric = fabric.Fabric(self.conn, '/redfish/v1/Fabrics/SAS',
                                    redfish_version='1.0.3')

    def test__parse_attributes(self):
        # | WHEN |
        self.fabric._parse_attributes(self.json_doc)
        # | THEN |
        self.assertEqual('1.0.3', self.fabric.redfish_version)
        self.assertEqual('SAS', self.fabric.identity)
        self.assertEqual('SAS Fabric', self.fabric.name)
        self.assertEqual('A SAS Fabric with redundant switches.',
                         self.fabric.description)
        self.assertEqual(sushy.Protocol.SAS,
                         self.fabric.fabric_type)
        self.assertEqual(sushy.State.ENABLED, self.fabric.status.state)
        self.assertEqual(sushy.Health.OK, self.fabric.status.health)

    def test_endpoints(self):
        # | GIVEN |
        with open('sushy/tests/unit/json_samples/'
                  'endpoint_collection.json') as f:
            endpoint_collection_return_value = json.load(f)

        with open('sushy/tests/unit/json_samples/'
                  'endpoint.json') as f:
            endpoint_return_value = json.load(f)

        self.conn.get.return_value.json.side_effect = [
            endpoint_collection_return_value, endpoint_return_value]

        # | WHEN |
        actual_endpoints = self.fabric.endpoints

        # | THEN |
        self.assertIsInstance(actual_endpoints,
                              endpoint.EndpointCollection)
        self.assertEqual(actual_endpoints.name, 'Endpoint Collection')

        member = actual_endpoints.get_member(
            '/redfish/v1/Fabrics/SAS/Endpoints/Drive1')

        self.assertEqual(member.name, "SAS Drive")
        self.assertEqual(member.endpoint_protocol, sushy.Protocol.SAS)

    def test_endpoints_on_refresh(self):
        # | GIVEN |
        with open('sushy/tests/unit/json_samples/'
                  'endpoint_collection.json') as f:
            self.conn.get.return_value.json.return_value = json.load(f)

        # | WHEN & THEN |
        endpts = self.fabric.endpoints
        self.assertIsInstance(endpts, endpoint.EndpointCollection)

        # On refreshing the fabric instance...
        with open('sushy/tests/unit/json_samples/fabric.json') as f:
            self.conn.get.return_value.json.return_value = json.loads(f.read())

        self.fabric.invalidate()
        self.fabric.refresh(force=False)

        # | WHEN & THEN |
        self.assertTrue(endpts._is_stale)

        # | GIVEN |
        with open('sushy/tests/unit/json_samples/'
                  'endpoint_collection.json') as f:
            self.conn.get.return_value.json.return_value = json.load(f)

        # | WHEN & THEN |
        self.assertIsInstance(self.fabric.endpoints,
                              endpoint.EndpointCollection)
        self.assertFalse(endpts._is_stale)


class FabricCollectionTestCase(base.TestCase):

    def setUp(self):
        super().setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/unit/json_samples/'
                  'fabric_collection.json') as f:
            self.conn.get.return_value.json.return_value = json.load(f)
        self.fabric = fabric.FabricCollection(
            self.conn, '/redfish/v1/Fabrics', '1.0.3', None)

    @mock.patch.object(fabric, 'Fabric', autospec=True)
    def test_get_member(self, fabric_mock):
        self.fabric.get_member('/redfish/v1/Fabrics/SAS1')
        fabric_mock.assert_called_once_with(
            self.fabric._conn, '/redfish/v1/Fabrics/SAS1',
            redfish_version=self.fabric.redfish_version, registries=None,
            root=self.fabric.root)

    @mock.patch.object(fabric, 'Fabric', autospec=True)
    def test_get_members(self, fabric_mock):
        members = self.fabric.get_members()
        calls = [
            mock.call(self.fabric._conn, '/redfish/v1/Fabrics/SAS1',
                      redfish_version=self.fabric.redfish_version,
                      registries=None, root=self.fabric.root),
            mock.call(self.fabric._conn, '/redfish/v1/Fabrics/SAS2',
                      redfish_version=self.fabric.redfish_version,
                      registries=None, root=self.fabric.root)
        ]
        fabric_mock.assert_has_calls(calls)
        self.assertIsInstance(members, list)
        self.assertEqual(2, len(members))



# Copyright (c) 2021 Dell Inc. or its subsidiaries.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
from unittest import mock

from oslotest.base import BaseTestCase

from sushy import exceptions
from sushy.oem.dell.resources.system import raid_service
from sushy import taskmonitor


class DellRaidService(BaseTestCase):

    def setUp(self):
        super().setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/oem/dell/unit/json_samples/'
                  'raid_service.json') as f:
            mock_response = self.conn.get.return_value
            mock_response.json.return_value = json.load(f)
            mock_response.status_code = 200

        self.mock_response = mock.Mock()
        self.mock_response.status_code = 202
        self.mock_response.headers = {
            'Location': '/redfish/v1/Managers/iDRAC.Embedded.1/Oem/Dell/Jobs/'
                        'JID_999888777666'
        }

        self.root = mock.Mock()

        self.raid_service = raid_service.DellRaidService(
            self.conn,
            '/redfish/v1/Systems/System.Embedded.1/Oem/Dell/'
            'DellRaidService', root=self.root
        )

    @mock.patch.object(raid_service.DellRaidService,
                       '_get_task_monitor_from_dell_job', autospec=True)
    def test_convert_to_raid(self, mock_get_task_mon):
        mock_task_mon = mock.Mock()
        mock_get_task_mon.return_value = mock_task_mon
        fqdds = ["Disk.Bay.0:Enclosure.Internal.0-1:RAID.Integrated.1-1",
                 "Disk.Bay.1:Enclosure.Internal.0-1:RAID.Integrated.1-1"]

        task_mon = self.raid_service.convert_to_raid(fqdds)

        self.conn.post.assert_called_once_with(
            '/redfish/v1/Systems/System.Embedded.1/Oem/Dell/DellRaidService/'
            'Actions/DellRaidService.ConvertToRAID',
            data={'PDArray': fqdds})
        self.assertEqual(mock_task_mon, task_mon)

    @mock.patch.object(raid_service.DellRaidService,
                       '_get_task_monitor_from_dell_job', autospec=True)
    def test_convert_to_nonraid(self, mock_get_task_mon):
        mock_task_mon = mock.Mock()
        mock_get_task_mon.return_value = mock_task_mon
        fqdds = ["Disk.Bay.0:Enclosure.Internal.0-1:RAID.Integrated.1-1",
                 "Disk.Bay.1:Enclosure.Internal.0-1:RAID.Integrated.1-1"]

        task_mon = self.raid_service.convert_to_nonraid(fqdds)

        self.conn.post.assert_called_once_with(
            '/redfish/v1/Systems/System.Embedded.1/Oem/Dell/DellRaidService/'
            'Actions/DellRaidService.ConvertToNonRAID',
            data={'PDArray': fqdds})
        self.assertEqual(mock_task_mon, task_mon)

    @mock.patch.object(raid_service.DellRaidService,
                       '_get_task_monitor_from_dell_job', autospec=True)
    def test_clear_foreign_config(self, mock_get_task_mon):
        mock_task_mon = mock.Mock()
        mock_get_task_mon.return_value = mock_task_mon

        result = self.raid_service.clear_foreign_config('RAID.Integrated.1-1')

        self.conn.post.assert_called_once_with(
            '/redfish/v1/Systems/System.Embedded.1/Oem/Dell/DellRaidService/'
            'Actions/DellRaidService.ClearForeignConfig',
            data={'TargetFQDD': 'RAID.Integrated.1-1'})
        self.assertEqual(mock_task_mon, result)

    def test_clear_foreign_config_no_config(self):
        mock_response = mock.Mock()
        mock_response.status_code = 400
        mock_response.json.return_value = {
            "error": {
                "@Message.ExtendedInfo": [
                    {
                        "Message": "No foreign configurations detected.",
                        "MessageArgs": [],
                        "MessageArgs@odata.count": 0,
                        "MessageId": "IDRAC.2.5.STOR018",
                        "RelatedProperties": [],
                        "RelatedProperties@odata.count": 0,
                        "Resolution": "If the only foreign drives present are "
                                      "in a secured locked state, run a "
                                      "secure erase operation on the drives "
                                      "to securely erase data or unlock these "
                                      "drives and retry the operation. "
                                      "Otherwise the operation was not "
                                      "successful because there are no "
                                      "foreign drives.",
                        "Severity": "Warning"
                    }
                ],
                "code": "Base.1.8.GeneralError",
                "message": "A general error has occurred. See ExtendedInfo "
                           "for more information"
            }
        }
        no_config_error = exceptions.BadRequestError(
            'POST', '/redfish/v1/Dell/Systems/System.Embedded.1/'
            'DellRaidService/Actions/DellRaidService.ClearForeignConfig',
            mock_response)
        self.conn.post.side_effect = no_config_error

        result = self.raid_service.clear_foreign_config('RAID.Integrated.1-1')

        self.assertIsNone(result)

    def test_clear_foreign_config_bad_request(self):
        mock_response = mock.Mock()
        mock_response.status_code = 400
        mock_response.json.return_value = {
            "error": {
                "@Message.ExtendedInfo": [
                    {
                        "Message": "Controller not found.",
                        "MessageArgs": [],
                        "MessageArgs@odata.count": 0,
                        "MessageId": "IDRAC.2.4.STOR030",
                        "RelatedProperties": [],
                        "RelatedProperties@odata.count": 0,
                        "Resolution": "Provide a valid controller FQDD (Fully "
                                      "Qualified Device Descriptor) and retry "
                                      "the operation.",
                        "Severity": "Warning"
                    }
                ],
                "code": "Base.1.7.GeneralError",
                "message": "A general error has occurred. See ExtendedInfo "
                           "for more information"
            }
        }
        no_config_error = exceptions.BadRequestError(
            'POST', '/redfish/v1/Dell/Systems/System.Embedded.1/'
            'DellRaidService/Actions/DellRaidService.ClearForeignConfig',
            mock_response)
        self.conn.post.side_effect = no_config_error

        self.assertRaises(exceptions.BadRequestError,
                          self.raid_service.clear_foreign_config,
                          'RAID.Integrated.999')

    def test__get_task_monitor_from_dell_job(self):
        mock_task1 = mock.Mock(identity='JID_111222333444',
                               path='/TaskService/Task/JID_111222333444')
        mock_task2 = mock.Mock(identity='JID_999888777666',
                               path='/TaskService/Task/JID_999888777666')
        mock_tasks = mock.Mock()
        mock_tasks.get_members.return_value = [mock_task1, mock_task2]
        self.root.get_task_service.return_value.tasks = mock_tasks

        task_mon = self.raid_service._get_task_monitor_from_dell_job(
            self.mock_response)

        self.assertIsInstance(task_mon, taskmonitor.TaskMonitor)
        self.assertEqual('/TaskService/Task/JID_999888777666',
                         task_mon.task_monitor_uri)

    def test__get_task_monitor_from_dell_job_location_missing(self):
        mock_response = mock.Mock()
        mock_response.status_code = 202
        mock_response.headers = {
            'Connection': 'Keep-Alive'
        }

        self.assertRaisesRegex(
            exceptions.ExtensionError,
            'does not include Location',
            self.raid_service._get_task_monitor_from_dell_job, mock_response)

    def test__get_task_monitor_from_dell_job_task_not_found(self):
        mock_task1 = mock.Mock(identity='JID_000000000000',
                               path='/TaskService/Task/JID_000000000000')
        mock_tasks = mock.Mock()
        mock_tasks.get_members.return_value = [mock_task1]
        self.root.get_task_service.return_value.tasks = mock_tasks

        self.assertRaisesRegex(
            exceptions.ExtensionError,
            'not find task by id',
            self.raid_service._get_task_monitor_from_dell_job,
            self.mock_response)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import time

import sushy

LOG = logging.getLogger(__name__)


def reboot_system(system):
    if system.power_state != sushy.POWER_STATE_OFF:
        system.reset_system(sushy.RESET_FORCE_OFF)
        LOG.info('Requested system power OFF')

    while system.power_state != sushy.POWER_STATE_OFF:
        time.sleep(30)
        system.refresh()

    LOG.info('System is powered OFF')

    system.reset_system(sushy.RESET_ON)

    LOG.info('Requested system power ON')

    while system.power_state != sushy.POWER_STATE_ON:
        time.sleep(30)
        system.refresh()

    LOG.info('System powered ON')



# Copyright (c) 2021 Dell Inc. or its subsidiaries.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

from sushy.oem.dell.resources.taskservice import constants as ts_cons
from sushy.resources import base
from sushy.resources.oem import base as oem_base

LOG = logging.getLogger(__name__)


class DellTaskExtension(oem_base.OEMResourceBase):
    """Dell OEM extension for DellJob type"""
    identity = base.Field('Id', required=True)

    name = base.Field('Name', required=True)

    description = base.Field('Description')

    completion_time = base.Field('CompletionTime')
    """Job completion time"""

    end_time = base.Field('EndTime')
    """End time of job

    Timestamp until when the service will wait for a job to complete.
    If a job does not complete within this time, it is killed and marked
    as failed. TIME_NA is a default value and implies EndTime is not
    applicable.
    """

    job_state = base.MappedField('JobState', ts_cons.JobState)
    """Job state"""

    job_type = base.MappedField('JobType', ts_cons.JobType)
    """Job type"""

    message = base.Field('Message')
    """The status message for job"""

    message_args = base.Field('MessageArgs')
    """Array of message arguments for message field"""

    message_id = base.Field('MessageId')
    """Message id for job"""

    percent_complete = base.Field('PercentComplete', adapter=int)
    """The percentage completion of job"""

    start_time = base.Field('StartTime')
    """Scheduled start time of job

    String that will contain a timestamp in Edm.DateTime format.
    TIME_NOW is a default value and implies apply pending configuration
    now.
    """


def get_extension(*args, **kwargs):
    return DellTaskExtension



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
from unittest import mock

from sushy import exceptions
from sushy.resources.system import constants
from sushy.resources.system import secure_boot
from sushy.resources.system import secure_boot_database
from sushy.tests.unit import base


class SecureBootTestCase(base.TestCase):

    def setUp(self):
        super().setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/unit/json_samples/secure_boot.json') as f:
            self.secure_boot_json = json.load(f)

        self.conn.get.return_value.json.return_value = self.secure_boot_json
        self.conn.get.return_value.headers = {'ETag': 'b26ae716a2c1f39f'}
        self.secure_boot = secure_boot.SecureBoot(
            self.conn, '/redfish/v1/Systems/437XR1138R2/SecureBoot',
            registries={}, redfish_version='1.1.0')

    def test__parse_attributes(self):
        self.secure_boot._parse_attributes(self.secure_boot_json)
        self.assertEqual('1.1.0', self.secure_boot.redfish_version)
        self.assertEqual('SecureBoot', self.secure_boot.identity)
        self.assertEqual('UEFI Secure Boot', self.secure_boot.name)
        self.assertIsNone(self.secure_boot.description)
        self.assertIs(False, self.secure_boot.enabled)
        self.assertEqual(constants.SecureBootCurrentBoot.DISABLED,
                         self.secure_boot.current_boot)
        self.assertEqual(constants.SecureBootMode.DEPLOYED,
                         self.secure_boot.mode)

    @mock.patch.object(secure_boot.LOG, 'warning', autospec=True)
    def test_get_allowed_reset_keys_values(self, mock_log):
        self.assertEqual({
            constants.SecureBootResetKeysType.RESET_ALL_KEYS_TO_DEFAULT,
            constants.SecureBootResetKeysType.DELETE_ALL_KEYS,
            constants.SecureBootResetKeysType.DELETE_PK
        }, self.secure_boot.get_allowed_reset_keys_values())
        self.assertFalse(mock_log.called)

    @mock.patch.object(secure_boot.LOG, 'warning', autospec=True)
    def test_get_allowed_reset_keys_values_no_values(self, mock_log):
        self.secure_boot._actions.reset_keys.allowed_values = None
        self.assertEqual({
            constants.SecureBootResetKeysType.RESET_ALL_KEYS_TO_DEFAULT,
            constants.SecureBootResetKeysType.DELETE_ALL_KEYS,
            constants.SecureBootResetKeysType.DELETE_PK
        }, self.secure_boot.get_allowed_reset_keys_values())
        self.assertTrue(mock_log.called)

    @mock.patch.object(secure_boot.LOG, 'warning', autospec=True)
    def test_get_allowed_reset_keys_values_custom_values(self, mock_log):
        self.secure_boot._actions.reset_keys.allowed_values = [
            'ResetAllKeysToDefault',
            'IamNotRedfishCompatible',
        ]
        self.assertEqual(
            {constants.SecureBootResetKeysType.RESET_ALL_KEYS_TO_DEFAULT},
            self.secure_boot.get_allowed_reset_keys_values())
        self.assertFalse(mock_log.called)

    def test_set_enabled(self):
        self.secure_boot.set_enabled(True)
        self.conn.patch.assert_called_once_with(
            '/redfish/v1/Systems/437XR1138R2/SecureBoot',
            data={'SecureBootEnable': True},
            etag='b26ae716a2c1f39f')

    def test_set_enabled_wrong_type(self):
        self.assertRaises(exceptions.InvalidParameterValueError,
                          self.secure_boot.set_enabled, 'banana')

    def test_reset_keys(self):
        self.secure_boot.reset_keys(
            constants.SecureBootResetKeysType.RESET_ALL_KEYS_TO_DEFAULT)
        self.conn.post.assert_called_once_with(
            '/redfish/v1/Systems/437XR1138R2/SecureBoot'
            '/Actions/SecureBoot.ResetKeys',
            data={'ResetKeysType': 'ResetAllKeysToDefault'})

    def test_reset_keys_wrong_value(self):
        self.assertRaises(exceptions.InvalidParameterValueError,
                          self.secure_boot.reset_keys, 'DeleteEverything')

    def test_databases(self):
        self.conn.get.return_value.json.reset_mock()

        with open('sushy/tests/unit/json_samples/'
                  'secure_boot_database_collection.json') as f:
            self.conn.get.return_value.json.return_value = json.load(f)

        result = self.secure_boot.databases
        self.assertIsInstance(
            result, secure_boot_database.SecureBootDatabaseCollection)
        self.conn.get.return_value.json.assert_called_once_with()

        self.conn.get.return_value.json.reset_mock()

        self.assertIs(result, self.secure_boot.databases)
        self.conn.get.return_value.json.assert_not_called()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# This is referred from Redfish standard schema.
# https://redfish.dmtf.org/schemas/v1/PCIeDevice.v1_19_0.json
# Per DMTF DSP0268_2025.2 Section 6.96 PCIeDevice 1.19.0
# Per DMTF DSP0268_2025.2 Section 6.97 PCIeFunction 1.6.0
import logging

from sushy import exceptions
from sushy.resources import base
from sushy.resources import common
from sushy.resources.system import constants as sys_cons
from sushy import utils

LOG = logging.getLogger(__name__)


class PCIeInterfaceField(base.CompositeField):
    """PCIe interface information for the device."""

    lanes_in_use = base.Field('LanesInUse', adapter=utils.int_or_none)
    """The number of PCIe lanes in use by this device."""

    max_lanes = base.Field('MaxLanes', adapter=utils.int_or_none)
    """The number of PCIe lanes supported by this device."""

    max_pcie_type = base.Field('MaxPCIeType')
    """The highest version of the PCIe specification supported by this
device."""
    pcie_type = base.Field('PCIeType')
    """The version of the PCIe specification in use by this device."""


class SlotField(base.CompositeField):
    """Slot information for the PCIe device."""

    lanes = base.Field('Lanes', adapter=utils.int_or_none)
    """The number of PCIe lanes supported by this slot."""

    location = base.Field('Location')
    """The location of the PCIe slot."""

    pcie_type = base.Field('PCIeType')
    """The PCIe specification this slot supports."""

    slot_type = base.MappedField('SlotType', sys_cons.SlotType)
    """The PCIe slot type."""

    hot_pluggable = base.Field('HotPluggable', adapter=bool)
    """An indication of whether this PCIe slot supports hotplug."""

    lane_splitting = base.MappedField('LaneSplitting',
                                      sys_cons.LaneSplittingType)
    """The lane splitting strategy used in the PCIe slot."""


class PCIeFunction(base.ResourceBase):
    """Represents a PCIe function associated within a PCIe device."""

    identity = base.Field('Id', required=True)
    """The PCIe function identity string"""

    name = base.Field('Name')
    """The PCIe function name"""

    description = base.Field('Description')
    """The PCIe function description"""

    function_id = base.Field('FunctionId', adapter=utils.int_or_none)
    """The PCIe function number within a given PCIe device."""

    function_type = base.MappedField('FunctionType', sys_cons.FunctionType)
    """The type of the PCIe function (Physical or Virtual)."""

    function_protocol = base.MappedField('FunctionProtocol',
                                         sys_cons.FunctionProtocol)
    """The PCIe function protocol (PCIe or CXL)."""

    device_class = base.MappedField('DeviceClass', sys_cons.DeviceClass)
    """The class for this PCIe function
    (NetworkController, MassStorageController, etc.)."""

    device_id = base.Field('DeviceId')
    """The Device ID of this PCIe function."""

    vendor_id = base.Field('VendorId')
    """The Vendor ID of this PCIe function."""

    subsystem_id = base.Field('SubsystemId')
    """The Subsystem ID of this PCIe function."""

    subsystem_vendor_id = base.Field('SubsystemVendorId')
    """The Subsystem Vendor ID of this PCIe function."""

    revision_id = base.Field('RevisionId')
    """The Revision ID of this PCIe function."""

    class_code = base.Field('ClassCode')
    """The Class Code of this PCIe function."""

    bus_number = base.Field('BusNumber')
    """The bus number of this PCIe function."""

    device_number = base.Field('DeviceNumber')
    """The device number of this PCIe function."""

    function_number = base.Field('FunctionNumber')
    """The function number of this PCIe function."""

    segment_number = base.Field('SegmentNumber')
    """The segment number of this PCIe function."""

    enabled = base.Field('Enabled', adapter=bool)
    """An indication of whether this PCIe device function is enabled."""

    status = common.StatusField('Status')
    """The status and health of the resource and its subordinate resources."""

    links = base.Field('Links')
    """Links to related resources."""


class PCIeFunctionCollection(base.ResourceCollectionBase):
    """Represents a collection of PCIe functions associated with a device."""

    @property
    def _resource_type(self):
        return PCIeFunction

    @property
    @utils.cache_it
    def function_count(self):
        """The number of PCIe functions in the collection.

        Returns the cached value until it (or its parent resource)
        is refreshed.
        """
        return len(self.get_members())


class PCIeDevice(base.ResourceBase):
    """Represents a PCIe device associated with a system."""

    identity = base.Field('Id', required=True)
    """The PCIe device identity string"""

    name = base.Field('Name')
    """The PCIe device name"""

    description = base.Field('Description')
    """The PCIe device description"""

    manufacturer = base.Field('Manufacturer')
    """The manufacturer of this PCIe device."""

    model = base.Field('Model')
    """The model number for the PCIe device."""

    serial_number = base.Field('SerialNumber')
    """The serial number for this PCIe device."""

    part_number = base.Field('PartNumber')
    """The part number for this PCIe device."""

    sku = base.Field('SKU')
    """The stock-keeping unit for this PCIe device."""

    device_type = base.MappedField('DeviceType', sys_cons.PCIeDeviceType)
    """The device type for this PCIe device."""

    firmware_version = base.Field('FirmwareVersion')
    """The version of firmware for this PCIe device."""

    asset_tag = base.Field('AssetTag')
    """The user-assigned asset tag for this PCIe device."""

    status = common.StatusField('Status')
    """The status and health of the resource and its subordinate resources."""

    pcie_interface = PCIeInterfaceField('PCIeInterface')
    """The PCIe interface details for this device."""

    slot = SlotField('Slot')
    """Information about the slot for this PCIe device."""

    links = base.Field('Links')
    """Links to related resources."""

    @property
    @utils.cache_it
    def pcie_functions(self):
        """Property to reference PCIeFunctionCollection instance

        Returns a PCIeFunctionCollection representing the PCIe functions
        associated with this PCIe device.

        :raises: MissingAttributeError if PCIeFunctions is not supported
        :returns: PCIeFunctionCollection instance
        """
        try:
            return PCIeFunctionCollection(
                self._conn,
                utils.get_sub_resource_path_by(self, "PCIeFunctions"),
                redfish_version=self.redfish_version,
                registries=self.registries, root=self.root)
        except exceptions.MissingAttributeError:
            # Return empty collection if PCIeFunctions is not supported
            return PCIeFunctionCollection(
                self._conn, "/empty",
                redfish_version=self.redfish_version,
                registries=self.registries, root=self.root)


class PCIeDeviceCollection(base.ResourceCollectionBase):
    @property
    def _resource_type(self):
        return PCIeDevice

    @property
    @utils.cache_it
    def device_count(self):
        """The number of PCIe devices in the collection.

        Returns the cached value until it (or its parent resource)
        is refreshed.
        """
        return len(self.get_members())

    def __init__(self, connector, path, redfish_version=None, registries=None,
                 root=None, embedded_data=None):
        if path == "/empty":
            self._conn = connector
            self._path = path
            self._json = {'Members': []}
            self.redfish_version = redfish_version
            self._registries = registries
            self._root = root
            self._is_stale = False
            return

        if embedded_data is not None:
            self._conn = connector
            self._path = path
            self._json = {'Members': embedded_data}
            self.redfish_version = redfish_version
            self._registries = registries
            self._root = root
            self._is_stale = False
            return

        super().__init__(connector, path, redfish_version, registries, root)

    @property
    def members_identities(self):
        if self._path == "/empty":
            return []
        if hasattr(self, '_json') and self._json and 'Members' in self._json:
            return tuple(m['@odata.id'] for m in self._json['Members']
                         if isinstance(m, dict) and '@odata.id' in m)
        return super().members_identities

    def get_members(self):
        if self._path == "/empty":
            return []

        # Handle embedded PCIeDevices case
        if hasattr(self, '_json') and self._json and 'Members' in self._json:
            device_objects = []
            for member in self._json['Members']:
                if isinstance(member, dict) and '@odata.id' in member:
                    device_path = member['@odata.id']
                    try:
                        # Fetch device data and create object
                        device_response = self._conn.get(device_path)
                        if device_response:
                            device_obj = PCIeDevice(
                                self._conn, device_path,
                                redfish_version=self.redfish_version,
                                registries=self._registries, root=self._root
                            )
                            device_obj._json = device_response.json()
                            device_objects.append(device_obj)
                    except Exception as e:
                        LOG.warning(
                            'Error fetching PCIe device at path %s: %s',
                            device_path, str(e))
                        continue  # Skip failed devices but continue processing
            return device_objects

        # Standard collection behavior
        return super().get_members()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# Values come from the Redfish ResourceBlock json-schema.
# https://redfish.dmtf.org/schemas/ResourceBlock.v1_3_0.json

import enum


class CompositionState(enum.Enum):
    COMPOSING = 'Composing'
    """Intermediate state indicating composition is in progress."""

    COMPOSED_AND_AVAILABLE = 'ComposedAndAvailable'
    """Indicates the Resource Block is currently participating in one or
    more compositions, and is available to be used in more compositions."""

    COMPOSED = 'Composed'
    """Final successful state of a Resource Block which has participated in
    composition."""

    UNUSED = 'Unused'
    """Indicates the Resource Block is free and can participate in
    composition."""

    FAILED = 'Failed'
    """The final composition resulted in failure and manual intervention may
    be required to fix it."""

    UNAVAILABLE = 'Unavailable'
    """Indicates the Resource Block has been made unavailable by the
    service, such as due to maintenance being performed on the Resource
    Block."""


# Backward compatibility
COMPOSITION_STATE_COMPOSING = CompositionState.COMPOSING
COMPOSITION_STATE_COMPOSED_AND_AVAILABLE = \
    CompositionState.COMPOSED_AND_AVAILABLE
COMPOSITION_STATE_COMPOSED = CompositionState.COMPOSED
COMPOSITION_STATE_UNUSED = CompositionState.UNUSED
COMPOSITION_STATE_FAILED = CompositionState.FAILED
COMPOSITION_STATE_UNAVAILABLE = CompositionState.UNAVAILABLE


class ResourceBlockType(enum.Enum):
    COMPUTE = 'Compute'
    """This Resource Block contains both Processor and Memory resources in a
    manner that creates a compute complex."""

    PROCESSOR = 'Processor'
    """This Resource Block contains Processor resources."""

    MEMORY = 'Memory'
    """This Resource Block contains Memory resources."""

    NETWORK = 'Network'
    """This Resource Block contains Network resources, such as Ethernet
    Interfaces."""

    STORAGE = 'Storage'
    """This Resource Block contains Storage resources, such as Storage and
    Simple Storage."""

    COMPUTER_SYSTEM = 'ComputerSystem'
    """This Resource Block contains ComputerSystem resources."""

    EXPANSION = 'Expansion'
    """This Resource Block is capable of changing over time based on its
    configuration.  Different types of devices within this Resource Block
    can be added and removed over time."""


# Backward compatibility
RESOURCE_BLOCK_TYPE_COMPUTE = ResourceBlockType.COMPUTE
RESOURCE_BLOCK_TYPE_PROCESSOR = ResourceBlockType.PROCESSOR
RESOURCE_BLOCK_TYPE_MEMORY = ResourceBlockType.MEMORY
RESOURCE_BLOCK_TYPE_NETWORK = ResourceBlockType.NETWORK
RESOURCE_BLOCK_TYPE_STORAGE = ResourceBlockType.STORAGE
RESOURCE_BLOCK_TYPE_COMPUTERSYSTEM = ResourceBlockType.COMPUTER_SYSTEM
RESOURCE_BLOCK_TYPE_EXPANSION = ResourceBlockType.EXPANSION



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# EventType constants
# http://redfish.dmtf.org/schemas/v1/Event.json#/definitions/EventType
# https://redfish.dmtf.org/schemas/v1/EventService.v1_0_6.json

import enum


class EventType(enum.Enum):

    STATUS_CHANGE = 'StatusChange'
    """The status of a resource has changed."""

    RESOURCE_UPDATED = 'ResourceUpdated'
    """A resource has been updated."""

    RESOURCE_ADDED = 'ResourceAdded'
    """A resource has been added."""

    RESOURCE_REMOVED = 'ResourceRemoved'
    """A resource has been removed."""

    ALERT = 'Alert'
    """A condition requires attention."""

    METRIC_REPORT = 'MetricReport'
    """The telemetry service is sending a metric report."""

    OTHER = 'Other'
    """Because EventType is deprecated as of Redfish Specification v1.6, the
    event is based on a registry or resource but not an EventType."""


# Backward compatibility
EVENT_TYPE_STATUS_CHANGE = EventType.STATUS_CHANGE
EVENT_TYPE_RESOURCE_UPDATED = EventType.RESOURCE_UPDATED
EVENT_TYPE_RESOURCE_ADDED = EventType.RESOURCE_ADDED
EVENT_TYPE_RESOURCE_REMOVED = EventType.RESOURCE_REMOVED
EVENT_TYPE_ALERT = EventType.ALERT
EVENT_TYPE_METRIC_REPORT = EventType.METRIC_REPORT
EVENT_TYPE_OTHER = EventType.OTHER



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# Values comes from the Redfish System json-schema:
# http://redfish.dmtf.org/schemas/v1/Resource.json and
# https://redfish.dmtf.org/schemas/v1/Settings.v1_3_3.json and
# https://redfish.dmtf.org/schemas/v1/Protocol.json

import enum

from sushy.resources.registry import constants as reg_cons
from sushy.resources.taskservice import constants as ts_cons


class Health(enum.Enum):
    """Health related constants."""
    OK = 'OK'
    """Normal."""

    WARNING = 'Warning'
    """A condition requires attention."""

    CRITICAL = 'Critical'
    """A critical condition requires immediate attention."""


# Backward compatibility
HEALTH_OK = Health.OK
HEALTH_WARNING = Health.WARNING
HEALTH_CRITICAL = Health.CRITICAL


class State(enum.Enum):
    """State related constants."""
    ENABLED = 'Enabled'
    """This function or resource is enabled."""

    DISABLED = 'Disabled'
    """This function or resource is disabled."""

    STANDBY_OFFLINE = 'StandbyOffline'
    """This function or resource is enabled but awaits an external action to
    activate it."""

    STANDBY_SPARE = 'StandbySpare'
    """This function or resource is part of a redundancy set and awaits a
    failover or other external action to activate it."""

    IN_TEST = 'InTest'
    """This function or resource is undergoing testing, or is in the process
    of capturing information for debugging."""

    STARTING = 'Starting'
    """This function or resource is starting."""

    ABSENT = 'Absent'
    """This function or resource is either not present or detected."""

    UNAVAILABLE_OFFLINE = 'UnavailableOffline'
    """This function or resource is present but cannot be used."""

    DEFERRING = 'Deferring'
    """The element does not process any commands but queues new requests."""

    QUIESCED = 'Quiesced'
    """The element is enabled but only processes a restricted set of
    commands."""

    UPDATING = 'Updating'
    """The element is updating and might be unavailable or degraded."""

    QUALIFIED = 'Qualified'
    """The element quality is within the acceptable range of operation."""


# Backward compatibility
STATE_ENABLED = State.ENABLED
STATE_DISABLED = State.DISABLED
STATE_ABSENT = State.ABSENT
STATE_STANDBYOFFLINE = State.STANDBY_OFFLINE
STATE_STANDBYSPARE = State.STANDBY_SPARE
STATE_INTEST = State.IN_TEST
STATE_STARTING = State.STARTING
STATE_UNAVAILABLEOFFLINE = State.UNAVAILABLE_OFFLINE
STATE_DEFERRING = State.DEFERRING
STATE_QUIESCED = State.QUIESCED
STATE_UPDATING = State.UPDATING


# Backward compatibility, the type moved to taskservice.constants
TASK_STATE_NEW = ts_cons.TaskState.NEW
TASK_STATE_STARTING = ts_cons.TaskState.STARTING
TASK_STATE_RUNNING = ts_cons.TaskState.RUNNING
TASK_STATE_SUSPENDED = ts_cons.TaskState.SUSPENDED
TASK_STATE_INTERRUPTED = ts_cons.TaskState.INTERRUPTED
TASK_STATE_PENDING = ts_cons.TaskState.PENDING
TASK_STATE_STOPPING = ts_cons.TaskState.STOPPING
TASK_STATE_COMPLETED = ts_cons.TaskState.COMPLETED
TASK_STATE_KILLED = ts_cons.TaskState.KILLED
TASK_STATE_EXCEPTION = ts_cons.TaskState.EXCEPTION
TASK_STATE_SERVICE = ts_cons.TaskState.SERVICE
TASK_STATE_CANCELLING = ts_cons.TaskState.CANCELLING
TASK_STATE_CANCELLED = ts_cons.TaskState.CANCELLED


# Backward compatibility, the type moved to registry.constants
PARAMTYPE_STRING = reg_cons.MessageParamType.STRING
PARAMTYPE_NUMBER = reg_cons.MessageParamType.NUMBER


# Backward compatibility (Severity is an alias of Health after 1.1.0)
Severity = Health
SEVERITY_OK = Severity.OK
SEVERITY_WARNING = Severity.WARNING
SEVERITY_CRITICAL = Severity.CRITICAL


class IndicatorLED(enum.Enum):
    """Indicator LED Constants"""

    LIT = 'Lit'
    """The Indicator LED is lit"""

    BLINKING = 'Blinking'
    """The Indicator LED is blinking"""

    OFF = 'Off'
    """The Indicator LED is off"""

    UNKNOWN = 'Unknown'
    """The state of the Indicator LED cannot be determine"""


# Backward compatibility
INDICATOR_LED_LIT = IndicatorLED.LIT
INDICATOR_LED_BLINKING = IndicatorLED.BLINKING
INDICATOR_LED_OFF = IndicatorLED.OFF
INDICATOR_LED_UNKNOWN = IndicatorLED.UNKNOWN


class PowerState(enum.Enum):
    """System PowerState constants"""

    ON = 'On'
    """The resource is powered on"""

    OFF = 'Off'
    """The resource is powered off, although some components may continue to
    have AUX power such as management controller"""

    POWERING_ON = 'PoweringOn'
    """A temporary state between Off and On. This temporary state can
    be very short"""

    POWERING_OFF = 'PoweringOff'
    """A temporary state between On and Off. The power off action can take
    time while the OS is in the shutdown process"""

    PAUSED = 'Paused'
    """The resource is paused."""


# Backward compatibility
POWER_STATE_ON = PowerState.ON
POWER_STATE_OFF = PowerState.OFF
POWER_STATE_POWERING_ON = PowerState.POWERING_ON
POWER_STATE_POWERING_OFF = PowerState.POWERING_OFF


class ResetType(enum.Enum):
    """Reset action constants"""
    ON = 'On'
    """Turn on the unit."""

    FORCE_OFF = 'ForceOff'
    """Turn off the unit immediately (non-graceful shutdown)."""

    GRACEFUL_SHUTDOWN = 'GracefulShutdown'
    """Shut down gracefully and power off."""

    GRACEFUL_RESTART = 'GracefulRestart'
    """Shut down gracefully and restart the system."""

    FORCE_RESTART = 'ForceRestart'
    """Shut down immediately and non-gracefully and restart the system."""

    NMI = 'Nmi'
    """Generate a diagnostic interrupt, which is usually an NMI on x86
    systems, to stop normal operations, complete diagnostic actions, and,
    typically, halt the system."""

    FORCE_ON = 'ForceOn'
    """Turn on the unit immediately."""

    PUSH_POWER_BUTTON = 'PushPowerButton'
    """Simulate the pressing of the physical power button on this unit."""

    POWER_CYCLE = 'PowerCycle'
    """Power cycle the unit.  Behaves like a full power removal, followed by
    a power restore to the resource."""

    SUSPEND = 'Suspend'
    """Write the state of the unit to disk before powering off.  This allows
    for the state to be restored when powered back on."""

    PAUSE = 'Pause'
    """Pause execution on the unit but do not remove power.  This is
    typically a feature of virtual machine hypervisors."""

    RESUME = 'Resume'
    """Resume execution on the paused unit.  This is typically a feature of
    virtual machine hypervisors."""


# Backward compatibility
RESET_TYPE_ON = ResetType.ON
RESET_TYPE_FORCE_OFF = ResetType.FORCE_OFF
RESET_TYPE_GRACEFUL_SHUTDOWN = ResetType.GRACEFUL_SHUTDOWN
RESET_TYPE_GRACEFUL_RESTART = ResetType.GRACEFUL_RESTART
RESET_TYPE_FORCE_RESTART = ResetType.FORCE_RESTART
RESET_TYPE_NMI = ResetType.NMI
RESET_TYPE_FORCE_ON = ResetType.FORCE_ON
RESET_TYPE_PUSH_POWER_BUTTON = ResetType.PUSH_POWER_BUTTON
RESET_TYPE_POWER_CYCLE = ResetType.POWER_CYCLE


class Protocol(enum.Enum):
    """Protocol type constants"""

    PCIe = 'PCIe'
    """PCI Express."""

    AHCI = 'AHCI'
    """Advanced Host Controller Interface (AHCI)."""

    UHCI = 'UHCI'
    """Universal Host Controller Interface (UHCI)."""

    SAS = 'SAS'
    """Serial Attached SCSI."""

    SATA = 'SATA'
    """Serial AT Attachment."""

    USB = 'USB'
    """Universal Serial Bus (USB)."""

    NVMe = 'NVMe'
    """Non-Volatile Memory Express (NVMe)."""

    FC = 'FC'
    """Fibre Channel."""

    iSCSI = 'iSCSI'
    """Internet SCSI."""

    FCoE = 'FCoE'
    """Fibre Channel over Ethernet (FCoE)."""

    FCP = 'FCP'
    """Fibre Channel Protocol for SCSI."""

    FICON = 'FICON'
    """FIbre CONnection (FICON)."""

    NVMe_OVER_FABRICS = 'NVMeOverFabrics'
    """NVMe over Fabrics."""

    SMB = 'SMB'
    """Server Message Block (SMB).  Also known as the Common Internet File
    System (CIFS)."""

    NFSv3 = 'NFSv3'
    """Network File System (NFS) version 3."""

    NFSv4 = 'NFSv4'
    """Network File System (NFS) version 4."""

    HTTP = 'HTTP'
    """Hypertext Transport Protocol (HTTP)."""

    HTTPS = 'HTTPS'
    """Hypertext Transfer Protocol Secure (HTTPS)."""

    FTP = 'FTP'
    """File Transfer Protocol (FTP)."""

    SFTP = 'SFTP'
    """SSH File Transfer Protocol (SFTP)."""

    iWARP = 'iWARP'
    """Internet Wide Area RDMA Protocol (iWARP)."""

    RoCE = 'RoCE'
    """RDMA over Converged Ethernet Protocol."""

    RoCEv2 = 'RoCEv2'
    """RDMA over Converged Ethernet Protocol Version 2."""

    I2C = 'I2C'
    """Inter-Integrated Circuit Bus."""

    TCP = 'TCP'
    """Transmission Control Protocol (TCP)."""

    UDP = 'UDP'
    """User Datagram Protocol (UDP)."""

    TFTP = 'TFTP'
    """Trivial File Transfer Protocol (TFTP)."""

    GEN_Z = 'GenZ'
    """GenZ."""

    MULTI_PROTOCOL = 'MultiProtocol'
    """Multiple Protocols."""

    INFINI_BAND = 'InfiniBand'
    """InfiniBand."""

    ETHERNET = 'Ethernet'
    """Ethernet."""

    NVLINK = 'NVLink'
    """NVLink."""

    OEM = 'OEM'
    """OEM-specific."""

    DISPLAY_PORT = 'DisplayPort'
    """DisplayPort."""

    HDMI = 'HDMI'
    """HDMI."""

    VGA = 'VGA'
    """VGA."""

    DVI = 'DVI'
    """DVI."""


# Backward compatibility

PROTOCOL_TYPE_AHCI = Protocol.AHCI
PROTOCOL_TYPE_FC = Protocol.FC
PROTOCOL_TYPE_FCP = Protocol.FCP
PROTOCOL_TYPE_FCoE = Protocol.FCoE
PROTOCOL_TYPE_FICON = Protocol.FICON
PROTOCOL_TYPE_FTP = Protocol.FTP
PROTOCOL_TYPE_HTTP = Protocol.HTTP
PROTOCOL_TYPE_HTTPS = Protocol.HTTPS
PROTOCOL_TYPE_I2C = Protocol.I2C
PROTOCOL_TYPE_NFSv3 = Protocol.NFSv3
PROTOCOL_TYPE_NFSv4 = Protocol.NFSv4
PROTOCOL_TYPE_NVMe = Protocol.NVMe
PROTOCOL_TYPE_NVMeOverFabrics = Protocol.NVMe_OVER_FABRICS
PROTOCOL_TYPE_OEM = Protocol.OEM
PROTOCOL_TYPE_PCIe = Protocol.PCIe
PROTOCOL_TYPE_RoCE = Protocol.RoCE
PROTOCOL_TYPE_RoCEv2 = Protocol.RoCEv2
PROTOCOL_TYPE_SAS = Protocol.SAS
PROTOCOL_TYPE_SATA = Protocol.SATA
PROTOCOL_TYPE_SFTP = Protocol.SFTP
PROTOCOL_TYPE_SMB = Protocol.SMB
PROTOCOL_TYPE_TFTP = Protocol.TFTP
PROTOCOL_TYPE_UHCI = Protocol.UHCI
PROTOCOL_TYPE_USB = Protocol.USB
PROTOCOL_TYPE_iSCSI = Protocol.iSCSI
PROTOCOL_TYPE_iWARP = Protocol.iWARP

# These values are not in the Protocol enum, using them is probably wrong!
PROTOCOL_TYPE_CIFS = Protocol.SMB
PROTOCOL_TYPE_NFS = Protocol.NFSv4
PROTOCOL_TYPE_SCP = Protocol.SFTP


class DurableNameFormat(enum.Enum):
    """Durable name format constants"""
    NAA = 'NAA'
    """The Name Address Authority (NAA) format."""

    iQN = 'iQN'
    """The iSCSI Qualified Name (iQN)."""

    FC_WWN = 'FC_WWN'
    """The Fibre Channel (FC) World Wide Name (WWN)."""

    UUID = 'UUID'
    """The Universally Unique Identifier (UUID)."""

    EUI = 'EUI'
    """The IEEE-defined 64-bit Extended Unique Identifier (EUI)."""

    NQN = 'NQN'
    """The NVMe Qualified Name (NQN)."""

    NSID = 'NSID'
    """The NVM Namespace Identifier (NSID)."""

    NGUID = 'NGUID'
    """The Namespace Globally Unique Identifier (NGUID)."""


# Backward compatibility
DURABLE_NAME_FORMAT_NAA = DurableNameFormat.NAA
DURABLE_NAME_FORMAT_iQN = DurableNameFormat.iQN
DURABLE_NAME_FORMAT_FC_WWN = DurableNameFormat.FC_WWN
DURABLE_NAME_FORMAT_UUID = DurableNameFormat.UUID
DURABLE_NAME_FORMAT_EUI = DurableNameFormat.EUI
DURABLE_NAME_FORMAT_NQN = DurableNameFormat.NQN
DURABLE_NAME_FORMAT_NSID = DurableNameFormat.NSID


class ApplyTime(enum.Enum):
    """Apply time constants"""

    IMMEDIATE = 'Immediate'
    """Apply immediately."""

    ON_RESET = 'OnReset'
    """Apply on a reset."""

    AT_MAINTENANCE_WINDOW_START = 'AtMaintenanceWindowStart'
    """Apply during a maintenance window as specified by an administrator."""

    IN_MAINTENANCE_WINDOW_ON_RESET = 'InMaintenanceWindowOnReset'
    """Apply after a reset but within maintenance window as specified by an
    administrator."""


# Backward compatibility
APPLY_TIME_IMMEDIATE = ApplyTime.IMMEDIATE
APPLY_TIME_ON_RESET = ApplyTime.ON_RESET
APPLY_TIME_MAINT_START = ApplyTime.AT_MAINTENANCE_WINDOW_START
APPLY_TIME_MAINT_RESET = ApplyTime.IN_MAINTENANCE_WINDOW_ON_RESET



# Copyright (c) 2020-2021 Dell Inc. or its subsidiaries.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

from sushy.oem.dell import constants
from sushy.resources import base

LOG = logging.getLogger(__name__)


class DellJobCollection(base.ResourceBase):

    _JOB_EXPAND = '?$expand=.($levels=1)'

    def __init__(self, connector, identity, redfish_version=None,
                 registries=None):
        """A class representing a DellJobCollection.

        :param connector: A Connector instance
        :param identity: The identity of the DellJobCollection resource
        :param redfish_version: The version of Redfish. Used to construct
            the object according to schema of the given version.
        :param registries: Dict of Redfish Message Registry objects to be
            used in any resource that needs registries to parse messages
        """
        super().__init__(
            connector, identity, redfish_version, registries)

    def get_unfinished_jobs(self):
        """Get the unfinished jobs.

        :returns: A list of unfinished jobs.
        """
        job_expand_uri = f'{self._path}{self._JOB_EXPAND}'
        unfinished_jobs = []
        LOG.debug('Getting unfinished jobs...')
        job_response = self._conn.get(job_expand_uri)
        data = job_response.json()
        for job in data['Members']:
            if job['JobState'] in constants.INCOMPLETE_JOB_STATES:
                unfinished_jobs.append(job['Id'])
        LOG.info('Got unfinished jobs')
        return unfinished_jobs



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

# Sushy Redfish Authentication Modes

import abc
import logging

from sushy import exceptions

LOG = logging.getLogger(__name__)


class AuthBase(metaclass=abc.ABCMeta):

    def __init__(self, username=None, password=None):
        """A class representing a base Sushy authentication mechanism

        :param username: User account with admin/server-profile
            access privilege.
        :param password: User account password.
        """
        self._username = username
        self._password = password
        self._root_resource = None
        self._connector = None

    def set_context(self, root_resource, connector):
        """Set the context of the authentication object.

        :param root_resource: Root sushy object
        :param connector: Connector for http connections
        """
        # Set the root resource, and connector to use
        # for normal operations.
        self._root_resource = root_resource
        self._connector = connector
        self._connector.set_auth(self)

    def authenticate(self):
        """Perform authentication.

        :raises: RuntimeError
        """
        if self._root_resource is None or self._connector is None:
            raise RuntimeError('_root_resource / _connector is missing. '
                               'Forgot to call set_context()?')
        self._do_authenticate()

    @abc.abstractmethod
    def _do_authenticate(self):
        """Method to establish a session to a Redfish controller.

        Needs to be implemented by extending auth class,
        because each authentication type will authenticate in its own way.
        """

    @abc.abstractmethod
    def can_refresh_session(self):
        """Method to assert if session based refresh can be done."""

    def close(self):
        """Shutdown Redfish authentication object

        Undoes whatever should be undone to cancel authenticated session.
        """

    def __enter__(self):
        """Allow object to be called with the 'with' statement."""
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        """Allow object to be called with the 'with' statement.

        Allow object to be called with the 'with' statement but
        also ensure we call close method on exit.
        """
        self.close()


class BasicAuth(AuthBase):
    """Basic Authentication class.

       This is a class used to encapsulate a basic authentication session.

       :param username: User account with admin/server-profile
           access privilege.
       :param password: User account password.
    """

    def _do_authenticate(self):
        """Attempts to establish a Basic Authentication Session.

        """
        LOG.debug('Setting basic authentication connector information.')
        self._connector.set_http_basic_auth(self._username, self._password)

    def can_refresh_session(self):
        """Method to assert if session based refresh can be done."""
        return False


class SessionAuth(AuthBase):
    """Session Authentication class.

       This is a class used to encapsulate a redfish session.
    """

    def __init__(self, username=None, password=None):
        """A class representing a Session Authentication object.

        :param username: User account with admin/server-profile access
             privilege.
        :param password: User account password.
        """
        self._session_key = None
        """Our Sessions Key"""
        self._session_resource_id = None
        """Our Sessions Unique Resource ID or URL"""
        self._session_auth_previously_successful = False
        """Our reminder for tracking if session auth has previously worked."""

        super().__init__(username, password)

    def get_session_key(self):
        """Returns the session key.

        :returns: The session key.
        """
        return self._session_key

    def get_session_resource_id(self):
        """Returns the session resource id.

        :returns: The session resource id.
        """
        return self._session_resource_id

    def _do_authenticate(self):
        """Establish a redfish session.

        :raises: MissingXAuthToken
        :raises: ConnectionError
        :raises: AccessError
        :raises: HTTPError
        """
        auth_token = None

        auth_token, session_uri = self._root_resource.create_session(
            self._username, self._password)
        # Record the session authentication data.
        self._session_key = auth_token
        self._session_resource_id = session_uri
        # Set flag so we know we've previously successfully achieved
        # session authentication in order to lockout possible fallback during
        # session refresh/renegotiation.
        self._session_auth_previously_successful = True
        self._connector.set_http_session_auth(auth_token)

    def can_refresh_session(self):
        """Method to assert if session based refresh can be done."""
        return (self._session_key is not None
                and self._session_resource_id is not None)

    def refresh_session(self):
        """Method to refresh a session to a Redfish controller.

        This method is called to create a new session after
        a session that has already been established
        has timed-out or expired.

        :raises: MissingXAuthToken
        :raises: ConnectionError
        :raises: AccessError
        :raises: HTTPError
        """
        self.reset_session_attrs()
        self._do_authenticate()

    def close(self):
        """Close the Redfish Session.

        Attempts to close an established RedfishSession by
        deleting it from the remote Redfish controller.
        """
        if self._session_resource_id is not None:
            try:
                self._connector.delete(self._session_resource_id)
            except (exceptions.AccessError,
                    exceptions.ServerSideError) as exc:
                LOG.warning('Received exception "%(exception)s" while '
                            'attempting to delete the active session: '
                            '%(session_id)s',
                            {'exception': exc,
                             'session_id': self._session_resource_id})
            self.reset_session_attrs()

    def reset_session_attrs(self):
        """Reset active session related attributes."""
        self._session_key = None
        self._session_resource_id = None
        # Requests session object data is merged with user submitted data
        # per https://requests.readthedocs.io/en/master/user/advanced/
        # so we need to clear data explicitly set on the session too.
        self._connector._session.auth = None
        if 'X-Auth-Token' in self._connector._session.headers:
            # Delete the token value that was saved to the session
            # as otherwise we would end up with a dictionary containing
            # a {'X-Auth-Token': null} being sent across to the remote
            # bmc.
            del self._connector._session.headers['X-Auth-Token']


class SessionOrBasicAuth(SessionAuth):

    def __init__(self, username=None, password=None):
        super().__init__(username, password)
        self.basic_auth = BasicAuth(username=username, password=password)

    def _fallback_to_basic_authentication(self):
        """Fallback to basic authentication."""
        self.reset_session_attrs()
        self.basic_auth.set_context(self._root_resource, self._connector)
        self.basic_auth.authenticate()

    def _do_authenticate(self):
        """Establish a RedfishSession.

        We will attempt to establish a redfish session. If we are unable
        to establish one, fallback to basic authentication.
        """
        try:
            # Attempt session based authentication
            super()._do_authenticate()
        except exceptions.AccessError as e:
            if (not self.can_refresh_session()
                    and not self._session_auth_previously_successful):
                # We should only try and fallback if we've not been able
                # to establish a session. If we had a session previously
                # and we're trying to fallback, something fishy is afoot.
                LOG.warning('Falling back to "Basic" authentication as '
                            'we have been unable to authenticate to the '
                            'BMC. Exception: %(exception)s.',
                            {'exception': e})
                self._fallback_to_basic_authentication()
            else:
                # We previously had session authentication, something
                # has changed which is far out of the ordinary.
                LOG.debug('Received exception "%(exception)s" while '
                          'attempting to re-establish a session. '
                          'Raising AccessError, as something fishy '
                          'has occurred and the BMC has suddenly '
                          'ceased to support Session based '
                          'authentication.',
                          {'exception': e})
                raise
        except exceptions.ConnectionError as e:
            # The reason to explicitly catch a connectivity failure is the
            # case where transitory connectivity failures can occur while
            # working to authenticate.
            LOG.debug('Encountered a connectivity failure while attempting '
                      'to authenticate. We will not explicitly fallback. '
                      'Error: %(exception)s',
                      {'exception': e})
            # Previously we would silently eat the failure as SushyError
            # and fallback as it is a general fault. Callers on direct
            # invocations through a connector _op method call can still
            # receive these exceptions, and applications like Ironic do
            # consider a client reuse disqualifier if there has been
            # a connection failure, so it is okay for us to fix the behavior
            # here.
            raise
        except exceptions.SushyError as e:
            LOG.debug('Received exception "%(exception)s" while '
                      'attempting to establish a session. '
                      'Falling back to basic authentication.',
                      {'exception': e})
            self._fallback_to_basic_authentication()

    def refresh_session(self):
        """Method to refresh a session to a Redfish controller.

        This method is called to create a new RedfishSession
        if we have previously established a RedfishSession and
        the previous session has timed-out or expired.
        If we did not previously have an established session,
        we simply return our BasicAuthentication requests.Session.
        """
        if self.can_refresh_session():
            super().refresh_session()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
# This is referred from Redfish standard schema.
# https://redfish.dmtf.org/schemas/v1/Port.v1_8_0.json

from sushy.resources import base
from sushy.resources import common
from sushy.resources.system.network import constants


class LLDPReceiveField(base.CompositeField):
    chassis_id = base.Field("ChassisId")
    """chassis ID received from the remote partner across this link."""

    port_id = base.Field("PortId")
    """A colon delimited string of hexadecimal octets identifying a port."""


class EthernetField(base.CompositeField):
    associated_mac_addresses = base.Field(
        "AssociatedMACAddresses", adapter=list
    )
    """The array of configured MAC addresses that are associated."""

    flow_control_configuration = base.MappedField(
        "FlowControlConfiguration", constants.FlowControl
    )
    """The locally configured 802.3x flow control setting."""

    flow_control_status = base.MappedField(
        "FlowControlStatus", constants.FlowControl
    )
    """The 802.3x flow control behavior negotiated with the link partner"""

    lldp_receive = LLDPReceiveField("LLDPReceive")
    """LLDP data being received on this link."""


class Port(base.ResourceBase):
    """This class adds the Port resource"""

    identity = base.Field("Id", required=True)
    """The Port identity string"""

    name = base.Field("Name", required=True)
    """The port name"""

    current_speed_gbps = base.Field("CurrentSpeedGbps", adapter=int)
    """The network port current link speed."""

    description = base.Field("Description")
    """The port description"""

    ethernet = EthernetField("Ethernet")
    """The Ethernet-specific properties of the port."""

    link_status = base.MappedField("LinkStatus", constants.PortLinkStatus)
    """The link status of the port."""

    port_id = base.Field("PortId")
    """The physical port id label for this port."""

    status = common.StatusField("Status")
    """The port status"""


class PortCollection(base.ResourceCollectionBase):
    @property
    def _resource_type(self):
        return Port



# Copyright (c) 2020-2021 Dell Inc. or its subsidiaries.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

from sushy.resources import base
from sushy.resources import common

LOG = logging.getLogger(__name__)


class ActionsField(base.CompositeField):
    remote_service_api_status = common.ActionField(
        "#DellLCService.GetRemoteServicesAPIStatus")


class DellLCService(base.ResourceBase):

    _actions = ActionsField('Actions')
    _OK_STATUS_CODE = 200
    _READY_STATUS = 'Ready'
    identity = base.Field('Id', required=True)

    def __init__(self, connector, identity, redfish_version=None,
                 registries=None):
        """A class representing a DellLCService.

        :param connector: A Connector instance
        :param identity: The identity of the DellLCService resource
        :param redfish_version: The version of Redfish. Used to construct
            the object according to schema of the given version.
        :param registries: Dict of Redfish Message Registry objects to be
            used in any resource that needs registries to parse messages.
        """
        super().__init__(
            connector, identity, redfish_version, registries)

    def _is_remote_service_api_status_ready(self, status_field):
        """Checks remote service status field

        :param status_field: Status field to check, e.g., LCStatus, RTStatus
        :returns: True if response returned and status field is Ready,
            otherwise False.
        """
        target_uri = self._actions.remote_service_api_status.target_uri
        response = self._conn.post(target_uri, data={})
        if response.status_code != self._OK_STATUS_CODE:
            return False
        data = response.json()
        return data[status_field] == self._READY_STATUS

    def is_idrac_ready(self):
        """Indicates if the iDRAC is ready to accept commands.

        :returns: A boolean value True/False based on remote service api status
            response.
        """
        LOG.debug('Checking to see if the iDRAC is ready...')
        return self._is_remote_service_api_status_ready('LCStatus')

    def is_realtime_ready(self):
        """Indicates if real-time operations are ready to be accepted.

        :returns: True if ready to accept real-time operations, otherwise
            false.
        """
        LOG.debug('Checking to see if the real-time operations are ready...')
        return self._is_remote_service_api_status_ready('RTStatus')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

from sushy.resources import base
from sushy.resources import common
from sushy.resources.oem import base as oem_base

LOG = logging.getLogger(__name__)


class ProductionLocationField(base.CompositeField):
    facility_name = base.Field('FacilityName')
    country = base.Field('Country')


class ContosoActionsField(base.CompositeField):
    reset = common.ResetActionField('#Contoso.Reset')


class FakeOEMSystemExtension(oem_base.OEMResourceBase):

    data_type = base.Field('@odata.type')
    name = base.Field('Name', required=True)
    production_location = ProductionLocationField('ProductionLocation')
    _actions = ContosoActionsField('Actions')

    def get_reset_system_path(self):
        return self._actions.reset.target_uri


def get_extension(*args, **kwargs):
    return FakeOEMSystemExtension



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# Values come from the Redfish schema:
# https://redfish.dmtf.org/schemas/v1/MessageRegistry.v1_4_2.json

import enum


class MessageParamType(enum.Enum):
    """Message Registry message parameter type related constants."""
    STRING = "string"
    NUMBER = "number"



# Copyright (c) 2021-2022 Dell Inc. or its subsidiaries.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import enum


class PhysicalDiskStateMode(enum.Enum):
    """Physical disk state mode constants"""

    RAID = 'RAID'
    """RAID physical disk state mode"""

    NONRAID = 'Non-RAID'
    """Non-RAID physical disk state mode"""


# For backward compatibility
PHYSICAL_DISK_STATE_MODE_RAID = PhysicalDiskStateMode.RAID
PHYSICAL_DISK_STATE_MODE_NONRAID = PhysicalDiskStateMode.NONRAID



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# This is referred from Redfish standard schema.
# http://redfish.dmtf.org/schemas/v1/Drive.v1_4_0.json

import logging

from sushy import exceptions
from sushy.resources import base
from sushy.resources import common
from sushy.resources import constants as res_cons
from sushy.resources.system.storage import volume
from sushy import utils

LOG = logging.getLogger(__name__)


class Drive(base.ResourceBase):
    """This class represents a disk drive or other physical storage medium."""

    block_size_bytes = base.Field('BlockSizeBytes', adapter=utils.int_or_none)
    """The size of the smallest addressable unit of this drive in bytes"""

    capacity_bytes = base.Field('CapacityBytes', adapter=utils.int_or_none)
    """The size in bytes of this Drive"""

    identifiers = common.IdentifiersListField('Identifiers', default=[])
    """The Durable names for the drive"""

    identity = base.Field('Id', required=True)
    """The Drive identity string"""

    indicator_led = base.MappedField('IndicatorLED', res_cons.IndicatorLED)
    """Whether the indicator LED is lit or off"""

    manufacturer = base.Field('Manufacturer')
    """This is the manufacturer of this drive"""

    media_type = base.Field('MediaType')
    """The type of media contained in this drive"""

    model = base.Field('Model')
    """This is the model number for the drive"""

    name = base.Field('Name')
    """The name of the resource"""

    part_number = base.Field('PartNumber')
    """The part number for this drive"""

    protocol = base.MappedField('Protocol', res_cons.Protocol)
    """Protocol this drive is using to communicate to the storage controller"""

    revision = base.Field("Revision")
    """The firmware/hardware version of the drive."""

    serial_number = base.Field('SerialNumber')
    """The serial number for this drive"""

    status = common.StatusField('Status')
    """This type describes the status and health of the drive"""

    @property
    @utils.cache_it
    def volumes(self):
        """A list of volumes that this drive is part of.

        Volumes that this drive either wholly or only partially contains.

        :raises: MissingAttributeError if '@odata.id' field is missing.
        :returns: A list of `Volume` instances
        """
        paths = utils.get_sub_resource_path_by(
            self, ["Links", "Volumes"], is_collection=True)

        return [volume.Volume(self._conn, path,
                              redfish_version=self.redfish_version,
                              registries=self.registries)
                for path in paths]

    def set_indicator_led(self, state):
        """Set IndicatorLED to the given state.

        :param state: Desired LED state, an IndicatorLED value.
        :raises: InvalidParameterValueError, if any information passed is
            invalid.
        """
        try:
            state = res_cons.IndicatorLED(state).value
        except ValueError:
            raise exceptions.InvalidParameterValueError(
                parameter='state', value=state,
                valid_values=' ,'.join(i.value for i in res_cons.IndicatorLED))

        etag = self._get_etag()
        data = {'IndicatorLED': state}

        self._conn.patch(self.path, data=data, etag=etag)
        self.invalidate()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import json
from unittest import mock

import sushy
from sushy.resources.certificateservice import certificate
from sushy.tests.unit import base


class CertificateTestCase(base.TestCase):

    def setUp(self):
        super().setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/unit/json_samples/certificate.json') as f:
            self.json_doc = json.load(f)

        self.conn.get.return_value.json.return_value = self.json_doc

        self.ident = \
            "/redfish/v1/Managers/BMC/NetworkProtocol/HTTPS/Certificates/1"
        self.cert = certificate.Certificate(
            self.conn, self.ident,
            redfish_version='1.0.2')

    def test__parse_attributes(self):
        self.cert._parse_attributes(self.json_doc)
        self.assertEqual('1.0.2', self.cert.redfish_version)
        self.assertEqual("HTTPS Certificate", self.cert.name)
        self.assertIn('BEGIN CERTIFICATE', self.cert.certificate_string)
        self.assertEqual(sushy.CertificateType.PEM,
                         self.cert.certificate_type)
        self.assertEqual("manager.contoso.org",
                         self.cert.issuer.common_name)
        self.assertEqual("manager.contoso.org",
                         self.cert.subject.common_name)
        self.assertEqual([sushy.KeyUsage.KEY_ENCIPHERMENT,
                          sushy.KeyUsage.SERVER_AUTHENTICATION],
                         self.cert.key_usage)
        self.assertEqual(datetime.datetime(2018, 9, 7, 13, 22, 5,
                                           tzinfo=datetime.timezone.utc),
                         self.cert.valid_not_before)
        self.assertEqual(datetime.datetime(2019, 9, 7, 13, 22, 5,
                                           tzinfo=datetime.timezone.utc),
                         self.cert.valid_not_after)
        self.assertTrue(self.cert.serial_number)
        self.assertTrue(self.cert.fingerprint)
        self.assertEqual("TPM_ALG_SHA1", self.cert.fingerprint_hash_algorithm)
        self.assertEqual("sha256WithRSAEncryption",
                         self.cert.signature_algorithm)


class CertificateCollectionTestCase(base.TestCase):

    def setUp(self):
        super().setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/unit/json_samples'
                  '/certificate_collection.json') as f:
            self.json_coll = json.load(f)

        with open('sushy/tests/unit/json_samples/certificate.json') as f:
            self.json_cert = json.load(f)

        self.conn.get.return_value.json.side_effect = [self.json_coll,
                                                       self.json_cert]

        self.ident = \
            "/redfish/v1/Managers/BMC/NetworkProtocol/HTTPS/Certificates"
        self.cert_coll = certificate.CertificateCollection(
            self.conn, self.ident,
            redfish_version='1.0.2')

    def test_get_member(self):
        member = f"{self.ident}/1"
        cert = self.cert_coll.get_member(member)
        self.assertEqual("1", cert.identity)
        self.conn.get.assert_has_calls([
            mock.call(path=self.ident),
            mock.call().json(),
            mock.call(path=member),
            mock.call().json(),
        ])

    def test_delete_member(self):
        member = f"{self.ident}/1"
        self.cert_coll.delete_member(member)
        self.conn.delete.assert_called_once_with(member)

    def test_create_member(self):
        # Another loading of collection is because of a refresh
        self.conn.get.return_value.json.side_effect = [self.json_coll,
                                                       self.json_cert]

        member = f"{self.ident}/1"
        self.conn.post.return_value.headers = {'Location': member}

        result = self.cert_coll.create_member(
            certificate_string='abcd',
            certificate_type=sushy.CertificateType.PEM)
        self.assertEqual("1", result.identity)

        self.conn.get.assert_has_calls([
            mock.call(path=self.ident),
            mock.call().json(),
            mock.call(path=self.ident),
            mock.call().json(),
            mock.call(path=member),
            mock.call().json(),
        ])
        self.conn.post.assert_called_once_with(
            self.ident, data={'CertificateString': 'abcd',
                              'CertificateType': 'PEM'})

    def test_create_member_no_location(self):
        self.conn.get.return_value.json.side_effect = [self.json_coll]
        self.conn.post.return_value.headers = {}

        result = self.cert_coll.create_member(
            certificate_string='abcd',
            certificate_type=sushy.CertificateType.PEM)
        self.assertIsNone(result)

        self.conn.get.assert_has_calls([
            mock.call(path=self.ident),
            mock.call().json(),
            mock.call(path=self.ident),
            mock.call().json(),
        ])
        self.conn.post.assert_called_once_with(
            self.ident, data={'CertificateString': 'abcd',
                              'CertificateType': 'PEM'})



# Copyright (c) 2020-2021 Dell Inc. or its subsidiaries.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
from unittest import mock

from oslotest.base import BaseTestCase

from sushy.oem.dell.resources.manager import constants as mgr_cons
from sushy.oem.dell.resources.manager import idrac_card_service


class DelliDRACCardServiceTestCase(BaseTestCase):

    def setUp(self):
        super().setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/oem/dell/unit/json_samples/'
                  'idrac_card_service.json') as f:
            mock_response = self.conn.get.return_value
            mock_response.json.return_value = json.load(f)
            mock_response.status_code = 200

        mock_response = self.conn.post.return_value
        mock_response.status_code = 202
        mock_response.headers.get.return_value = '1'
        self.idrac_card_service = idrac_card_service.DelliDRACCardService(
            self.conn, '/redfish/v1/Dell/Managers/iDRAC.Embedded.1')

    def test_reset_idrac(self):
        self.idrac_card_service.reset_idrac()
        target_uri = ('/redfish/v1/Dell/Managers/iDRAC.Embedded.1'
                      '/DelliDRACCardService'
                      '/Actions/DelliDRACCardService.iDRACReset')
        self.conn.post.assert_called_once_with(target_uri,
                                               data={'Force': 'Graceful'})

    def test_get_allowed_reset_idrac_values(self):
        expected_values = {mgr_cons.ResetType.GRACEFUL,
                           mgr_cons.ResetType.FORCE}
        allowed_values = \
            self.idrac_card_service.get_allowed_reset_idrac_values()
        self.assertEqual(expected_values, allowed_values)

    def test_get_allowed_reset_idrac_values_not_provided_by_idrac(self):
        idrac_service_json = self.idrac_card_service.json
        base_property = '#DelliDRACCardService.iDRACReset'
        remove_property = 'Force@Redfish.AllowableValues'
        idrac_service_json['Actions'][base_property].pop(remove_property)
        expected_values = {mgr_cons.ResetType.GRACEFUL,
                           mgr_cons.ResetType.FORCE}
        allowed_values = \
            self.idrac_card_service.get_allowed_reset_idrac_values()
        self.assertEqual(expected_values, allowed_values)



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import json
from unittest import mock

from sushy.resources.registry import attribute_registry
from sushy.tests.unit import base


class AttributeRegistryTestCase(base.TestCase):

    def setUp(self):
        super().setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/unit/json_samples/'
                  'bios_attribute_registry.json') as f:
            self.json_doc = json.load(f)

        self.conn.get.return_value.json.return_value = self.json_doc

        self.registry = attribute_registry.AttributeRegistry(
            self.conn, '/redfish/v1/Test/Bios/BiosRegistry',
            redfish_version='1.0.2')

    def test__parse_attributes(self):
        self.registry._parse_attributes(self.json_doc)
        self.assertEqual('BiosAttributeRegistryP89.v1_0_0',
                         self.registry.identity)
        self.assertEqual('BIOS Attribute Registry', self.registry.name)
        self.assertEqual('en', self.registry.language)
        self.assertEqual('This is a test of BIOS Attribute Registry',
                         self.registry.description)
        self.assertEqual('1.0.0', self.registry.registry_version)
        self.assertEqual('VendorA', self.registry.owning_entity)
        self.assertEqual([{'FirmwareVersion': '2.0', 'ProductName': 'Ultra 4',
                          'SystemId': 'URGR8'}],
                         self.registry.supported_systems)
        attributes = self.registry.registry_entries.attributes[0]
        self.assertEqual('SystemModelName', attributes.name)
        self.assertEqual('System Model Name', attributes.display_name)
        self.assertEqual(True, attributes.immutable)
        self.assertEqual(True, attributes.read_only)
        self.assertEqual('String', attributes.attribute_type)

        attributes = self.registry.registry_entries.attributes[1]
        self.assertEqual('ProcVirtualization', attributes.name)
        self.assertEqual('Virtualization Technology', attributes.display_name)
        self.assertEqual(False, attributes.immutable)
        self.assertEqual(False, attributes.read_only)
        self.assertEqual(True, attributes.reset_required)
        self.assertEqual('Enumeration', attributes.attribute_type)
        self.assertEqual([{'ValueDisplayName': 'Enabled',
                           'ValueName': 'Enabled'},
                          {'ValueDisplayName': 'Disabled',
                           'ValueName': 'Disabled'}],
                         attributes.allowable_values)

    def test__parse_attributes_zt(self):
        self.conn = mock.Mock()
        with open('sushy/tests/unit/json_samples/'
                  'bios_attribute_registry_zthardware.json') as f:
            self.json_doc = json.load(f)

        self.conn.get.return_value.json.return_value = self.json_doc
        self.registry = attribute_registry.AttributeRegistry(
            self.conn,
            '/redfish/v1/Registries/'
            'BiosAttributeRegistryProt0.208.208.0.json',
            redfish_version='1.0.2')

        self.registry._parse_attributes(self.json_doc)
        self.assertEqual('BiosAttributeRegistryProt0.208.208.0',
                         self.registry.identity)
        self.assertEqual('Prot0 BIOS Attribute Registry', self.registry.name)
        self.assertEqual('en-US', self.registry.language)
        self.assertEqual('This registry defines a representation of '
                         'BIOS Attribute instances',
                         self.registry.description)
        self.assertEqual('208.208.0', self.registry.registry_version)
        self.assertEqual('AMI', self.registry.owning_entity)

        attributes = self.registry.registry_entries.attributes[0]
        self.assertEqual('TCG003', attributes.name)
        self.assertEqual('TPM SUPPORT', attributes.display_name)
        self.assertEqual('Enable', attributes.default_value)
        self.assertIsNone(attributes.immutable)
        self.assertEqual(False, attributes.read_only)
        self.assertEqual('Enumeration', attributes.attribute_type)
        self.assertEqual(True, attributes.reset_required)
        self.assertEqual([{'ValueDisplayName': 'Disable',
                           'ValueName': 'Disable'},
                          {'ValueDisplayName': 'Enable',
                           'ValueName': 'Enable'}],
                         attributes.allowable_values)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# This is referred from Redfish standard schema.
# http://redfish.dmtf.org/schemas/v1/SimpleStorage.v1_2_0.json

import logging

from sushy.resources import base
from sushy.resources import common
from sushy.resources import constants as res_cons
from sushy import utils


LOG = logging.getLogger(__name__)


class DeviceListField(base.ListField):
    """The storage device/s associated with SimpleStorage."""

    name = base.Field('Name', required=True)
    """The name of the storage device"""

    capacity_bytes = base.Field('CapacityBytes', adapter=utils.int_or_none)
    """The size of the storage device."""

    status = common.StatusField('Status')
    """Describes the status and health of a storage device."""


class SimpleStorage(base.ResourceBase):
    """This class represents a simple storage.

    It represents the properties of a storage controller and its
    directly-attached devices. A storage device can be a disk drive or optical
    media device.
    """

    identity = base.Field('Id', required=True)
    """The SimpleStorage identity string"""

    name = base.Field('Name')
    """The name of the resource"""

    devices = DeviceListField('Devices', default=[])
    """The storage devices associated with this resource."""


class SimpleStorageCollection(base.ResourceCollectionBase):
    """Represents a collection of simple storage associated with system."""

    @property
    def _resource_type(self):
        return SimpleStorage

    @property
    @utils.cache_it
    def disks_sizes_bytes(self):
        """Sizes of each Disk in bytes in SimpleStorage collection resource.

        Returns the list of cached values until it (or its parent resource)
        is refreshed.
        """
        return sorted(device.capacity_bytes
                      for simpl_stor in self.get_members()
                      for device in simpl_stor.devices
                      if (device.status.state == res_cons.State.ENABLED
                          and device.capacity_bytes is not None))

    @property
    def max_size_bytes(self):
        """Max size available (in bytes) among all enabled Disk resources.

        Returns the cached value until it (or its parent resource) is
        refreshed.
        """
        return utils.max_safe(self.disks_sizes_bytes)



# Copyright (c) 2021 Dell Inc. or its subsidiaries.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

from sushy import exceptions
from sushy.oem.dell import constants
from sushy.resources import base
from sushy.resources import common
from sushy import taskmonitor

LOG = logging.getLogger(__name__)


class ActionsField(base.CompositeField):
    convert_to_raid = common.ActionField("#DellRaidService.ConvertToRAID")
    convert_to_nonraid = common.ActionField(
        "#DellRaidService.ConvertToNonRAID")
    clear_foreign_config = common.ActionField(
        "#DellRaidService.ClearForeignConfig")


class DellRaidService(base.ResourceBase):
    identity = base.Field('Id', required=True)
    _actions = ActionsField('Actions')

    def __init__(self, connector, identity, redfish_version=None,
                 registries=None, root=None):
        """A class representing a DellRaidService.

        :param connector: A Connector instance
        :param identity: The identity of the DellRaidService resource
        :param redfish_version: The version of Redfish. Used to
            construct the object according to schema of the given
            version.
        :param registries: Dict of Redfish Message Registry objects to
            be used in any resource that needs registries to parse
            messages.
        :param root: Sushy root object. Empty for Sushy root itself.
        """
        super().__init__(
            connector, identity, redfish_version=redfish_version,
            registries=registries, root=root)

    def convert_to_raid(self, physical_disk_fqdds):
        """Converts physical disks to a state usable for RAID

        :param physical_disk_fqdds: An array of FQDDs where each
            identifies a physical drive.
        :returns: Sushy's TaskMonitor instance for TaskService task
        """
        target_uri = self._actions.convert_to_raid.target_uri
        payload = {'PDArray': physical_disk_fqdds}
        response = self._conn.post(target_uri, data=payload)
        task_monitor = self._get_task_monitor_from_dell_job(response)
        LOG.info('Converting to RAID mode: %s', physical_disk_fqdds)
        return task_monitor

    def convert_to_nonraid(self, physical_disk_fqdds):
        """Converts physical disks to non-RAID state.

        :param physical_disk_fqdds: An array of FQDDs where each
            identifies a physical drive.
        :returns: Sushy's TaskMonitor instance for TaskService task
        """
        target_uri = self._actions.convert_to_nonraid.target_uri
        payload = {'PDArray': physical_disk_fqdds}
        response = self._conn.post(target_uri, data=payload)
        task_monitor = self._get_task_monitor_from_dell_job(response)
        LOG.info('Converting to non-RAID mode: %s', physical_disk_fqdds)
        return task_monitor

    def clear_foreign_config(self, controller_fqdd):
        """Clears foreign configuration

        Prepares any foreign physical disks for inclusion in the local
        configuration

        :param controller_fqdd: FQDD of controller to clear foreign
            config

        :returns: Sushy's TaskMonitor instance for TaskService task if
            there are foreign drives to clear, otherwise None.
        """
        target_uri = self._actions.clear_foreign_config.target_uri
        payload = {'TargetFQDD': controller_fqdd}

        try:
            response = self._conn.post(target_uri, data=payload)
        except exceptions.BadRequestError as ex:
            # Check if failed for no foreign drives
            errors = ex.body and ex.body.get('@Message.ExtendedInfo') or []

            no_foreign_conf = [x for x in errors if constants.NO_FOREIGN_CONFIG
                               in x.get('MessageId')]

            if len(no_foreign_conf) == 0:
                raise ex
            else:
                LOG.debug('%s: %s', no_foreign_conf[0].get('Message'),
                          controller_fqdd)
                return

        task_mon = self._get_task_monitor_from_dell_job(response)
        LOG.info('Clearing foreign config: %s', controller_fqdd)
        return task_mon

    def _get_task_monitor_from_dell_job(self, response):
        """From OEM job response returns generic Task monitor

        :param response: Response from OEM job
        :returns: Sushy's TaskMonitor instance for TaskService task
        """
        location = response.headers.get('Location')
        if not location:
            raise exceptions.ExtensionError(
                error=f'Response {response.url} does not include Location '
                      'in header')

        task_id = location.split('/')[-1]
        task = None

        for t in self.root.get_task_service().tasks.get_members():
            if t.identity == task_id:
                task = t
                break

        if not task:
            raise exceptions.ExtensionError(
                error=f"Did not find task by id {task_id}")

        return taskmonitor.TaskMonitor(
            self._conn, task.path,
            redfish_version=self.redfish_version,
            registries=self.registries)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# Values come from the Redfish Fabric json-schema:
# https://redfish.dmtf.org/schemas/v1/Fabric.v1_2_2.json
# https://redfish.dmtf.org/schemas/v1/Endpoint.v1_6_1.json

import enum


class EntityRole(enum.Enum):
    """Entity role constants"""

    INITIATOR = 'Initiator'
    """The entity sends commands, messages, or other types of requests to
    other entities on the fabric, but cannot receive commands from other
    entities."""

    TARGET = 'Target'
    """The entity receives commands, messages, or other types of requests
    from other entities on the fabric, but cannot send commands to other
    entities."""

    BOTH = 'Both'
    """The entity can both send and receive commands, messages, and other
    requests to or from other entities on the fabric."""


# Backward compatibility
ENTITY_ROLE_INITIATOR = EntityRole.INITIATOR
ENTITY_ROLE_TARGET = EntityRole.TARGET
ENTITY_ROLE_BOTH = EntityRole.BOTH


class EntityType(enum.Enum):
    """Entity type constants"""

    STORAGE_INITIATOR = 'StorageInitiator'
    """The entity is a storage initiator."""

    ROOT_COMPLEX = 'RootComplex'
    """The entity is a PCI(e) root complex."""

    NETWORK_CONTROLLER = 'NetworkController'
    """The entity is a network controller."""

    DRIVE = 'Drive'
    """The entity is a drive."""

    STORAGE_EXPANDER = 'StorageExpander'
    """The entity is a storage expander."""

    DISPLAY_CONTROLLER = 'DisplayController'
    """The entity is a display controller."""

    BRIDGE = 'Bridge'
    """The entity is a PCI(e) bridge."""

    PROCESSOR = 'Processor'
    """The entity is a processor."""

    VOLUME = 'Volume'
    """The entity is a volume."""

    ACCELERATION_FUNCTION = 'AccelerationFunction'
    """The entity is an acceleration function realized through a device,
    such as an FPGA."""

    MEDIA_CONTROLLER = 'MediaController'
    """The entity is a media controller."""

    MEMORY_CHUNK = 'MemoryChunk'
    """The entity is a memory chunk."""

    SWITCH = 'Switch'
    """The entity is a switch, not an expander.  Use `Expander` for
    expanders."""

    FABRIC_BRIDGE = 'FabricBridge'
    """The entity is a fabric bridge."""

    MANAGER = 'Manager'
    """The entity is a manager."""

    STORAGE_SUBSYSTEM = 'StorageSubsystem'
    """The entity is a storage subsystem."""


# Backward compatibility
ENTITY_TYPE_STORAGE_INITIATOR = EntityType.STORAGE_INITIATOR
ENTITY_TYPE_ROOT_COMPLEX = EntityType.ROOT_COMPLEX
ENTITY_TYPE_NETWORK_CONTROLLER = EntityType.NETWORK_CONTROLLER
ENTITY_TYPE_DRIVE = EntityType.DRIVE
ENTITY_TYPE_STORAGE_EXPANDER = EntityType.STORAGE_EXPANDER
ENTITY_TYPE_DISPLAY_CONTROLLER = EntityType.DISPLAY_CONTROLLER
ENTITY_TYPE_PCI_BRIDGE = EntityType.BRIDGE
ENTITY_TYPE_PROCESSOR = EntityType.PROCESSOR
ENTITY_TYPE_VOLUME = EntityType.VOLUME
